#include "pekon.h"

struct sheets *sctl;

void paletteInit(void)
{
	static unsigned char rgb[16 * 3] = {
		0x00, 0x00, 0x00,
		0xff, 0x00, 0x00,
		0x00, 0xff, 0x00,
		0xff, 0xff, 0x00,
		0x00, 0x00, 0xff,
		0xff, 0x00, 0xff,
		0x00, 0xff, 0xff,
		0xff, 0xff, 0xff,
		0xc6, 0xc6, 0xc6,
		0x84, 0x00, 0x00,
		0x00, 0x84, 0x00,
		0x84, 0x84, 0x00,
		0x00, 0x00, 0x84,
		0x84, 0x00, 0x84,
		0x00, 0x84, 0x84,
		0x84, 0x84, 0x84
	};
	paletteSet(rgb, 0, 15);
	return;
}

void paletteSet(unsigned char *rgb, int start, int end)
{
	int i, e;
	e = eflagsLoad();
	cli();
	out8(0x03c8, start);
	for (i = start; i <= end; i++) {
		out8(0x03c9, rgb[0] / 4);
		out8(0x03c9, rgb[1] / 4);
		out8(0x03c9, rgb[2] / 4);
		rgb += 3;
	}
	eflagsStore(e);
	return;
}

void shtInit(struct memory *mem)
{
	int i;
	sctl = (struct sheets *) memAlloc(mem, sizeof(struct sheets));
	sctl->top = -1;
	sctl->map = (unsigned char *) memAlloc(mem, 320 * 200);
	for (i = 0; i < SHEETS; i++) {
		sctl->sheet[i].flag = 0;
	}
	return;
}

struct sheet *shtAlloc(void)
{
	struct sheet *sht;
	int i;
	for (i = 0; i < SHEETS; i++) {
		if (sctl->sheet[i].flag == 0) {
			sht = &sctl->sheet[i];
			sht->flag = SHTUSE;
			sht->height = -1;
			return sht;
		}
	}
	return 0;
}

void shtSet(struct sheet *sht, unsigned char *buf, int sizex, int sizey, int cinv)
{
	sht->buf = buf;
	sht->sizex = sizex;
	sht->sizey = sizey;
	sht->cinv = cinv;
	return;
}

void shtUpdown(struct sheet *sht, int height)
{
	int h, old = sht->height;
	if (height > sctl->top + 1) {
		height = sctl->top + 1;
	} else if (height < -1) {
		height = -1;
	}
	sht->height = height;
	if (old > height) {
		if (height >= 0) {
			for (h = old; h > height; h--) {
				sctl->sht[h] = sctl->sht[h - 1];
				sctl->sht[h]->height = h;
			}
			sctl->sht[height] = sht;
			shtRefreshmap(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, height + 1);
			shtRefresh0(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, height + 1, old);
		} else {
			if (sctl->top > old) {
				for (h = old; h < sctl->top; h++) {
					sctl->sht[h] = sctl->sht[h + 1];
					sctl->sht[h]->height = h;
				}
			}
			sctl->top--;
			shtRefreshmap(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, 0);
			shtRefresh0(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, 0, old - 1);
		}
	} else if (old < height) {
		if (old >= 0) {
			for (h = old; h < height; h++) {
				sctl->sht[h] = sctl->sht[h + 1];
				sctl->sht[h]->height = h;
			}
			sctl->sht[height] = sht;
		} else {
			for (h = sctl->top; h >= height; h--) {
				sctl->sht[h + 1] = sctl->sht[h];
				sctl->sht[h + 1]->height = h + 1;
			}
			sctl->sht[height] = sht;
			sctl->top++;
		}
		shtRefreshmap(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, height);
		shtRefresh0(sht->x, sht->y, sht->x + sht->sizex, sht->y + sht->sizey, height, height);
	}
	return;
}

void shtRefresh(struct sheet *sht, int x0, int y0, int x1, int y1)
{
	if (sht->height >= 0) {
		shtRefresh0(sht->x + x0, sht->y + y0, sht->x + x1, sht->y + y1, sht->height, sht->height);
	}
	return;
}

void shtRefresh0(int vx0, int vy0, int vx1, int vy1, int h0, int h1)
{
	int h, bx, by, vx, vy, bx0, by0, bx1, by1;
	unsigned char *buf, sid, *vram = (char *) 0xa0000;
	struct sheet *sht;
	if (vx0 < 0) { vx0 = 0; }
	if (vy0 < 0) { vy0 = 0; }
	if (vx1 > 320) { vx1 = 320; }
	if (vy1 > 200) { vy1 = 200; }
	for (h = h0; h <= h1; h++) {
		sht = sctl->sht[h];
		buf = sht->buf;
		sid = sht - sctl->sheet;
		bx0 = vx0 - sht->x;
		by0 = vy0 - sht->y;
		bx1 = vx1 - sht->x;
		by1 = vy1 - sht->y;
		if (bx0 < 0) { bx0 = 0; }
		if (by0 < 0) { by0 = 0; }
		if (bx1 > sht->sizex) { bx1 = sht->sizex; }
		if (by1 > sht->sizey) { by1 = sht->sizey; }
		for (by = by0; by < by1; by++) {
			vy = sht->y + by;
			for (bx = bx0; bx < bx1; bx++) {
				vx = sht->x + bx;
				if (sctl->map[vy * 320 + vx] == sid) {
					vram[vy * 320 + vx] = buf[by * sht->sizex + bx];
				}
			}
		}
	}
	return;
}

void shtRefreshmap(int vx0, int vy0, int vx1, int vy1, int h0)
{
	int h, bx, by, vx, vy, bx0, by0, bx1, by1;
	unsigned char *buf, sid;
	struct sheet *sht;
	if (vx0 < 0) { vx0 = 0; }
	if (vy0 < 0) { vy0 = 0; }
	if (vx1 > 320) { vx1 = 320; }
	if (vy1 > 200) { vy1 = 200; }
	for (h = h0; h <= sctl->top; h++) {
		sht = sctl->sht[h];
		buf = sht->buf;
		sid = sht - sctl->sheet;
		bx0 = vx0 - sht->x;
		by0 = vy0 - sht->y;
		bx1 = vx1 - sht->x;
		by1 = vy1 - sht->y;
		if (bx0 < 0) { bx0 = 0; }
		if (by0 < 0) { by0 = 0; }
		if (bx1 > sht->sizex) { bx1 = sht->sizex; }
		if (by1 > sht->sizey) { by1 = sht->sizey; }
		for (by = by0; by < by1; by++) {
			vy = sht->y + by;
			for (bx = bx0; bx < bx1; bx++) {
				vx = sht->x + bx;
				if (buf[by * sht->sizex + bx] != sht->cinv) {
					sctl->map[vy * 320 + vx] = sid;
				}
			}
		}
	}
	return;
}

void shtSlide(struct sheet *sht, int x, int y)
{
	int ox = sht->x, oy = sht->y;
	sht->x = x;
	sht->y = y;
	if (sht->height >= 0) {
		shtRefreshmap(ox, oy, ox + sht->sizex, oy + sht->sizey, 0);
		shtRefreshmap(x, y, x + sht->sizex, y + sht->sizey, sht->height);
		shtRefresh0(ox, oy, ox + sht->sizex, oy + sht->sizey, 0, sht->height - 1);
		shtRefresh0(x, y, x + sht->sizex, y + sht->sizey, sht->height, sht->height);
	}
	return;
}

void shtFree(struct sheet *sht)
{
	if (sht->height >= 0) {
		shtUpdown(sht, -1);
	}
	sht->flag = 0;
	return;
}

void mouseCursor(struct sheet *sht)
{
	static char cursor[16][11] = {
		"*          ",
		"**         ",
		"***        ",
		"****       ",
		"*****      ",
		"******     ",
		"*******    ",
		"********   ",
		"*********  ",
		"********** ",
		"***********",
		"*****      ",
		"****       ",
		"***        ",
		"**         ",
		"*          "
	};
	int x, y;
	for (y = 0; y < 16; y++) {
		for (x = 0; x < 11; x++) {
			if (cursor[y][x] == '*') {
				sht->buf[y * sht->sizex + x] = _000000;
			} else {
				sht->buf[y * sht->sizex + x] = sht->cinv;
			}
		}
	}
	return;
}

void boxfill(struct sheet *sht, unsigned char c, int x0, int y0, int x1, int y1)
{
	int x, y;
	for (y = y0; y <= y1; y++) {
		for (x = x0; x <= x1; x++) {
			sht->buf[y * sht->sizex + x] = c;
		}
	}
	return;
}

void putfont(struct sheet *sht, unsigned char c, int x, int y, char *font)
{
	int i;
	char *p, d;
	for (i = 0; i < 16; i++) {
		p = sht->buf + (y + i) * sht->sizex + x;
		d = font[i];
		if ((d & 0x80) != 0) { p[0] = c; }
		if ((d & 0x40) != 0) { p[1] = c; }
		if ((d & 0x20) != 0) { p[2] = c; }
		if ((d & 0x10) != 0) { p[3] = c; }
		if ((d & 0x08) != 0) { p[4] = c; }
		if ((d & 0x04) != 0) { p[5] = c; }
		if ((d & 0x02) != 0) { p[6] = c; }
		if ((d & 0x01) != 0) { p[7] = c; }
	}
	return;
}

void putfonts(struct sheet *sht, int c, int x, int y, unsigned char *s)
{
	extern char hankaku[4096];
	for (; *s != 0x00; s++) {
		putfont(sht, c, x, y, hankaku + *s * 16);
		x += 8;
	}
	return;
}

void boxframe(struct sheet *sht, int c, int x0, int y0, int x1, int y1)
{
	boxfill(sht, c, x0, y0, x1, y0);
	boxfill(sht, c, x0, y0, x0, y1);
	boxfill(sht, c, x0, y1, x1, y1);
	boxfill(sht, c, x1, y0, x1, y1);
	return;
}

void winMake(struct sheet *sht, char *s, int act)
{
	boxframe(sht, _000000, 0, 0, sht->sizex - 1, sht->sizey - 1);
	boxfill(sht, _FFFFFF, 1, 21, sht->sizex - 2, sht->sizey - 2);
	wtitleMake(sht, s, act);
	return;
}

void shtPuts(struct sheet *sht, int bc, int c, int x, int y, char *s)
{
	int l;
	for (l = 0; s[l] != 0; l++) { }
	boxfill(sht, bc, x, y, x + l * 8 - 1, y + 15);
	putfonts(sht, c, x, y, s);
	shtRefresh(sht, x, y, x + l * 8, y + 16);
	return;
}

void wtitleMake(struct sheet *sht, char *s, int act)
{
	char bc;
	if (act != 0) {
		bc = _C6C6C6;
	} else {
		bc = _FFFFFF;
	}
	boxfill(sht, bc, 1, 1, sht->sizex - 2, 19);
	putfonts(sht, _000000, 24, 2, s);
	boxfill(sht, _000000, 0, 20, sht->sizex - 1, 20);
	boxframe(sht, _000000, 4, 4, 16, 16);
	return;
}
