#include "pekon.h"

struct fifo *hfifo;
struct timers *tctl;

void gdtidtInit(void)
{
	struct segdesc *gdt = (struct segdesc *) 0x00270000;
	struct gatedesc *idt = (struct gatedesc *) 0x0026f800;
	int i;
	for (i = 0; i < 8192; i++) {
		segdescSet(gdt + i, 0, 0, 0);
	}
	segdescSet(gdt + 1, 0xffffffff, 0x00000000, SEGRW);
	segdescSet(gdt + 2, PEKONLIMIT,  PEKONADDR, SEGRX);
	gdtrLoad(0xffff, 0x00270000);

	for (i = 0; i < 256; i++) {
		gatedescSet(idt + i, 0, 0, 0);
	}
	gatedescSet(idt + 0x0c, (int) int0cAsm, 2 * 8, GATEINT);
	gatedescSet(idt + 0x0d, (int) int0dAsm, 2 * 8, GATEINT);
	gatedescSet(idt + 0x20, (int) int20Asm, 2 * 8, GATEINT);
	gatedescSet(idt + 0x21, (int) int21Asm, 2 * 8, GATEINT);
	gatedescSet(idt + 0x2c, (int) int2cAsm, 2 * 8, GATEINT);
	gatedescSet(idt + 0x40, (int)   apiAsm, 2 * 8, GATEINT + 0x60);
	idtrLoad(0x07ff, 0x0026f800);
	return;
}

void segdescSet(struct segdesc *sd, unsigned int limit, unsigned int base, int ar)
{
	if (limit > 0x000fffff) {
		ar |= 0x8000;	/* G-bit */
		limit >>= 12;
	}
	sd->limitLow = limit & 0xffff;
	sd->baseLow = base & 0xffff;
	sd->baseMid = (base >> 16) & 0xff;
	sd->ar = ar & 0xff;
	sd->limitHigh = ((limit >> 16) & 0x0f) | ((ar >> 8) & 0xf0);
	sd->baseHigh = (base >> 24) & 0xff;
	return;
}

void gatedescSet(struct gatedesc *gd, unsigned int offset, int sel, int ar)
{
	gd->offsetLow = offset & 0xffff;
	gd->sel = sel;
	gd->dwCount = (ar >> 8) & 0xff;
	gd->ar = ar & 0xff;
	gd->offsetHigh = (offset >> 16) & 0xffff;
	return;
}

void picInit(void)
{
	out8(0x0021, 0xff);
	out8(0x00a1, 0xff);
	out8(0x0020, 0x11);
	out8(0x0021, 0x20);
	out8(0x0021, 1 << 2);
	out8(0x0021, 0x01);
	out8(0x00a0, 0x11);
	out8(0x00a1, 0x28);
	out8(0x00a1, 0x02);
	out8(0x00a1, 0x01);
	out8(0x0021, 0xfb);		/* 11111011 PIC1 */
	out8(0x00a1, 0xff);		/* 11111111 none */
	return;
}

void int21(int *esp)
{
	int i;
	out8(0x0020, 0x61);
	i = in8(0x0060);
	fifoPut(hfifo, 256 + i);
	return;
}

void int2c(int *esp)
{
	int i;
	out8(0x00a0, 0x64);
	out8(0x0020, 0x62);
	i = in8(0x0060);
	fifoPut(hfifo, 512 + i);
	return;
}

void kbdReady(void)
{
	for (;;) {
		if ((in8(0x0064) & 0x02) == 0) {
			return;
		}
	}
}

void keyInit(struct fifo *fifo)
{
	hfifo = fifo;
	kbdReady();
	out8(0x0064, 0x60);
	kbdReady();
	out8(0x0060, 0x47);
	return;
}

void mouseEnable(struct mouse *mouse)
{
	mouse->phase = 0;
	kbdReady();
	out8(0x0064, 0xd4);
	kbdReady();
	out8(0x0060, 0xf4);
	return;
}

int mouseDecode(struct mouse *mouse, unsigned char data)
{
	if (mouse->phase == 0) {
		if (data == 0xfa) {
			mouse->phase = 1;
		}
		return 0;
	} else if (mouse->phase == 1) {
		if ((data & 0xc8) == 0x08) {
			mouse->buf[0] = data;
			mouse->phase = 2;
		}
		return 0;
	} else if (mouse->phase == 2) {
		mouse->buf[1] = data;
		mouse->phase = 3;
		return 0;
	} else if (mouse->phase == 3) {
		mouse->buf[2] = data;
		mouse->phase = 1;
		mouse->btn = mouse->buf[0] & 0x07;
		mouse->x = mouse->buf[1];
		mouse->y = mouse->buf[2];
		if ((mouse->buf[0] & 0x10) != 0) {
			mouse->x |= 0xffffff00;
		}
		if ((mouse->buf[0] & 0x20) != 0) {
			mouse->y |= 0xffffff00;
		}
		mouse->y = -mouse->y;
		return 1;
	}
	return -1;
}

void timersInit(struct memory *mem)
{
	struct timer *t;
	int i;
	tctl = (struct timers *) memAlloc(mem, sizeof(struct timers));
	tctl->count = 0;
	for (i = 0; i < TIMERS; i++) {
		tctl->timer[i].flag = 0;
	}
	t = timerAlloc();
	t->tout = 0xffffffff;
	t->flag = TUSING;
	t->next = 0;
	tctl->t0 = t;
	out8(0x0043, 0x34);
	out8(0x0040, 0x9c);
	out8(0x0040, 0x2e);
	return;
}

void int20(int *esp)
{
	extern struct timer *ttask;
	char ts = 0;
	out8(0x0020, 0x60);
	tctl->count++;
	if (tctl->t0->tout > tctl->count) {
		return;
	}
	for (;;) {
		if (tctl->t0->tout > tctl->count) {
			break;
		}
		tctl->t0->flag = TALLOC;
		if (tctl->t0 != ttask) {
			fifoPut(tctl->t0->fifo, tctl->t0->data);
		} else {
			ts = 1;
		}
		tctl->t0 = tctl->t0->next;
	}
	if (ts != 0) {
		taskSwitch();
	}
	return;
}

struct timer *timerAlloc(void)
{
	int i;
	for (i = 0; i < TIMERS; i++) {
		if (tctl->timer[i].flag == 0) {
			tctl->timer[i].flag = TALLOC;
			return &tctl->timer[i];
		}
	}
	return 0;
}

void timerFree(struct timer *timer)
{
	timer->flag = 0;
	return;
}

void timerInit(struct timer *timer, struct fifo *fifo, int data)
{
	timer->fifo = fifo;
	timer->data = data;
	return;
}

void timerSet(struct timer *timer, unsigned int tout)
{
	struct timer *t, *s;
	int e;
	e = eflagsLoad();
	cli();
	timer->tout = tctl->count + tout;
	if (timer->flag == TUSING) {
		eflagsStore(e);
		return;
	}
	timer->flag = TUSING;
	t = tctl->t0;
	if (timer->tout <= t->tout) {
		tctl->t0 = timer;
		timer->next = t;
		eflagsStore(e);
		return;
	}
	for (;;) {
		s = t;
		t = t->next;
		if (timer->tout <= t->tout) {
			s->next = timer;
			timer->next = t;
			eflagsStore(e);
			return;
		}
	}
}

int *int0d(int *esp)
{
	struct task *task = taskNow();
	cli();
	fifoPut(hfifo, 0x0d);
	fifoPut(hfifo, (int) task);
	fifoPut(hfifo, esp[11]);
	sti();
	return &(task->tss.esp0);
}

int *int0c(int *esp)
{
	struct task *task = taskNow();
	cli();
	fifoPut(hfifo, 0x0c);
	fifoPut(hfifo, (int) task);
	fifoPut(hfifo, esp[11]);
	sti();
	return &(task->tss.esp0);
}
