#include "pekon.h"

struct tasks *tasks;
struct timer *ttask;

struct task *taskInit(struct memory *mem)
{
	struct segdesc *gdt = (struct segdesc *) 0x00270000;
	struct task *task, *idle;
	int i;
	tasks = (struct tasks *) memAlloc(mem, sizeof(struct tasks));
	for (i = 0; i < TASKS * TLEVELS; i++) {
		tasks->task[i].flag = 0;
		tasks->task[i].sel = (3 + i) * 8;
		segdescSet(gdt + 3 + i, 103, (int) &tasks->task[i].tss, TSS);
	}
	for (i = 0; i < TLEVELS; i++) {
		tasks->level[i].running = 0;
		tasks->level[i].now = 0;
	}
	task = taskAlloc();
	task->flag = 2;
	task->pri = 2;
	task->lv = 1;
	taskAdd(task);
	taskSwitch0();
	trLoad(task->sel);
	ttask = timerAlloc();
	timerSet(ttask, 2);

	idle = taskAlloc();
	idle->tss.esp = memAlloc(mem, 64 * 1024) + 64 * 1024;
	idle->tss.eip = (int) &taskIdle;
	idle->tss.es = 1 * 8;
	idle->tss.cs = 2 * 8;
	idle->tss.ss = 1 * 8;
	idle->tss.ds = 1 * 8;
	idle->tss.fs = 1 * 8;
	idle->tss.gs = 1 * 8;
	taskRun(idle, TLEVELS - 1, 1);

	return task;
}

void taskSwitch(void)
{
	struct tlevel *tl = &tasks->level[tasks->now];
	struct task *newt, *nowt = tl->task[tl->now];
	tl->now++;
	if (tl->now == tl->running) {
		tl->now = 0;
	}
	if (tasks->lvchan != 0) {
		taskSwitch0();
		tl = &tasks->level[tasks->now];
	}
	newt = tl->task[tl->now];
	timerSet(ttask, newt->pri);
	if (newt != nowt) {
		farjmp(0, newt->sel);
	}
	return;
}

struct task *taskAlloc(void)
{
	struct task *task;
	int i;
	for (i = 0; i < TASKS; i++) {
		if (tasks->task[i].flag == 0) {
			task = &tasks->task[i];
			task->flag = 1;
			task->tss.eflags = 0x00000202;	/* IF */
			task->tss.eax = 0;
			task->tss.ecx = 0;
			task->tss.edx = 0;
			task->tss.ebx = 0;
			task->tss.ebp = 0;
			task->tss.esi = 0;
			task->tss.edi = 0;
			task->tss.es = 0;
			task->tss.ds = 0;
			task->tss.fs = 0;
			task->tss.gs = 0;
			task->tss.ldtr = 0;
			task->tss.iomap = 0x40000000;
			return task;
		}
	}
	return 0;
}

void taskRun(struct task *task, int lv, int pri)
{
	if (lv < 0) {
		lv = task->lv;
	}
	if (pri > 0) {
		task->pri = pri;
	}
	if (task->flag == 2 && task->lv != lv) {
		taskRemove(task);
	}
	if (task->flag != 2) {
		task->lv = lv;
		taskAdd(task);
	}
	tasks->lvchan = 1;
	return;
}

void taskSleep(struct task *task)
{
	struct task *nt;
	if (task->flag == 2) {
		nt = taskNow();
		taskRemove(task);
		if (task == nt) {
			taskSwitch0();
			nt = taskNow();
			farjmp(0, nt->sel);
		}
	}
	return;
}

struct task *taskNow(void)
{
	struct tlevel *tl = &tasks->level[tasks->now];
	return tl->task[tl->now];
}

void taskAdd(struct task *task)
{
	struct tlevel *tl = &tasks->level[task->lv];
	tl->task[tl->running] = task;
	tl->running++;
	task->flag = 2;
	return;
}

void taskRemove(struct task *task)
{
	struct tlevel *tl = &tasks->level[task->lv];
	int i;
	for (i = 0; i < tl->running; i++) {
		if (tl->task[i] == task) {
			break;
		}
	}
	tl->running--;
	if (i < tl->now) {
		tl->now--;
	}
	if (tl->now >= tl->running) {
		tl->now = 0;
	}
	task->flag = 1;
	for (; i < tl->running; i++) {
		tl->task[i] = tl->task[i + 1];
	}
	return;
}

void taskSwitch0(void)
{
	int i;
	for (i = 0; i < TLEVELS; i++) {
		if (tasks->level[i].running > 0) {
			break;
		}
	}
	tasks->now = i;
	tasks->lvchan = 0;
	return;
}

void taskIdle(void)
{
	for (;;) {
		hlt();
	}
}
